//! Time-series features highly dependent on the cadence

mod duration;
pub use duration::Duration;

mod maximum_time_interval;
pub use maximum_time_interval::MaximumTimeInterval;

mod minimum_time_interval;
pub use minimum_time_interval::MinimumTimeInterval;

mod observation_count;
pub use observation_count::ObservationCount;

mod time_mean;
pub use time_mean::TimeMean;

mod time_standard_deviation;
pub use time_standard_deviation::TimeStandardDeviation;
